/* $Id: quant.c,v 1.7 1998/02/20 21:14:54 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "quant.h"


/*------------------------- QUANTIZATION --------------------------*/


/* 
   Function:
   ---------
   initQuantSingleStage - Initialization of the single-stage quantizer for
     a given value.

   Arguments:
   ----------
   quantState *state - Pointer to the state data structure.
   int initialValue - Value which is to be quantized.

   Return Value:
   -------------
   <None>

   Description:
   ------------
   This must be called prior to single-stage quantization. A seperate
   state structure must be kept for each value that is quantized in parallel.
   Single-stage quantization is just successive calls to quantSingleStage.
   For each value,  need only be called once, before the first call to 
   quantSingleStage.
*/

void initQuantSingleStage(quantState *state, int initialVal)
{
  state->residualValue = initialVal;
  state->QR            = 0;
  state->fromReduced   = 0;
  state->fromDeadZone  = 1;
}

/* 
   Function:
   ---------
   quantSingleStage - Single-stage quantizer.

   Arguments:
   ----------
   Q     - Quantization value. Represents desired quantization level size.
   state - State of quantizer.

   Return Value:
   -------------
   New Q index.

   Description:
   ------------
   initQuantSingleStage must be called prior to using this function the
   first time for a given value. It will compute the new quantization index
   based on the current state associated with the value.
*/

int quantSingleStage(int Q, quantState *state)
{
  int refLevs;    /* how many refinement levels in new stage */
  int QIndex;     /* new quantization index to return for this stage */

  /*--------------- INITIAL QUANTIZATION STAGE -------------------*/
  if (state->QR==0)
  {
    QIndex = state->residualValue/Q;

    /* update state */
    if (QIndex)
      state->residualValue = abs(state->residualValue) - (abs(QIndex)*Q);
    state->QR            = Q;
    state->fromReduced   = 0;
    state->fromDeadZone  = (QIndex==0);

    return QIndex;
  }


  /*--------------- NON-INITIAL QUANTIZATION STAGES -------------------*/

  /* get the number of refinement levels from lastQUsed state */
  refLevs = IROUND(state->QR,Q);

  /* Catch condition where there's no refinement being done.
     State information is not changed.
  */
  if (refLevs<=1)
    QIndex = 0;
  else
  {
    int inDeadZone;  /* are we still in the dead zone */
    int lastQUsed;   /* The "real" Q value last used */
    int val;         /* value of number to be quantized */
    int lastLevSize; /* Size of quantization level used last */
    int newQUsed;
    int excess;

    /* Initialize the last quant value used */
    lastQUsed = state->QR;
    
    /* update new Q value state */
    state->QR = newQUsed = ICEIL(lastQUsed,refLevs);
    
    /* Get last level size */
    lastLevSize = lastQUsed-(state->fromReduced);

    /* check if a reduced level can span the last level */
    if (refLevs*(newQUsed-1) >= lastLevSize)
    {
      --newQUsed;

      /* might overshoot (?) but can't reduce anymore */
      excess=0;
#if 1
      if (lastLevSize-refLevs*newQUsed)
	fprintf(stderr,"Excess in reduced partition\n");
#endif
    }
    else
      /* get excess (overshoot) */
      excess=lastLevSize-refLevs*newQUsed;

    /* Set dead zone indicator */
    inDeadZone = state->fromDeadZone;
    
    /* Set value of leftovers from last pass */
    val=state->residualValue;

    /*--- Calculate QIndex. Update residualValue and fromReduced states ---*/
    if (excess==0)
    {
      QIndex = val/newQUsed;
      state->fromReduced = (newQUsed < state->QR) ? 1 : 0;
      
      if (QIndex)
	state->residualValue -= QIndex*newQUsed;
    }
    else
    {
      int reducedParStart; /* Where the reduced partition starts (magnitude)
			    */
      int reducedIdx;

      reducedParStart = newQUsed*(refLevs+excess);
      if (abs(val) >= reducedParStart)
      {
	state->fromReduced = 1;
	QIndex = SGN(state->residualValue)*(refLevs+excess);
	state->residualValue -= QIndex*newQUsed;

	--newQUsed;
	reducedIdx = 
	  SGN(state->residualValue) * (abs(val)-reducedParStart) / newQUsed;
	QIndex += reducedIdx;
	state->residualValue -= reducedIdx*newQUsed;
      }
      else
      {
	state->fromReduced = 0;
	QIndex = val/newQUsed;
	state->residualValue -= QIndex*newQUsed;
      }
    }
    
    if (inDeadZone && QIndex)
    {
      /* We have the sign info so all residual values from here on in are 
	 positive */
      state->residualValue = abs(state->residualValue);
      state->fromDeadZone=0;
    }
  }

  return QIndex;
}

/* 
   Function:
   ---------
   quant - Complete quantization of a value.

   Arguments:
   ----------
   int *QVals - Pointer to array where quantized values are to be stored.
   int val - The value to be quantized.
   int numQ - The number of quantization stages.
   int *QList - The quantization values to be used for quantization. These
     will be modified if not multiples of one another.

   Return Value:
   -------------
   <none>

   Description:
   ------------
   quant will put the sequence of quantized values for the passed value,
   val, in the array pointed to by QVals. The quantization is based on the
   numQ Q values in QList.
*/

void quant(int *QVals, int val, int numQ, int *QList)
{
  int curIdx;
  quantState state;

  /* initialize state */
  initQuantSingleStage(&state, val);
  for (curIdx = 0; curIdx < numQ; ++curIdx)
    QVals[curIdx]=quantSingleStage(QList[curIdx], &state);
}


/*--------------- INVERSE QUANTIZATION --------------------------*/

/* 
   Function:
   ---------
   initInvQuantSingleStage - Initialization of the single-stage inverse
     quantizer for a given Q index.

   Arguments:
   ----------
   quantState *state - Pointer to the state data structure.
   int Q - initial Q value.

   Return Value:
   -------------
   <None>

   Description:
   ------------
   This must be called prior to single-stage inverse quantization. A seperate
   state structure must be kept for each value that is quantized in parallel.
   Single-stage inverse quantization is just successive calls to 
   invQuantSingleStage. For each value,  need only be called once, before
   the first call to invQuantSingleStage.
*/

void initInvQuantSingleStage(quantState *state)
{
  state->residualValue = 0;
  state->QR            = 0;
  state->fromReduced   = 0;
  state->fromDeadZone  = 1;
}

/* Middle of quantization level */
#ifdef _NEW_IQUANT_VAL_
#define QLEVMID(q) (((q)-1)/2)
#else
#define QLEVMID(q) (((q))/2)
#endif

/* Mapping of start of quantization level, sign, and quantization level size
   to specific value.
*/
#define GETVAL(start, sgn, q) ((start) \
			       ? ((start) + (sgn)*QLEVMID((q))) \
			       : 0)

/* 
   Function:
   ---------
   invQuantSingleStage - Single-stage inverse quantizer.

   Arguments:
   ----------
   QIndex - Quantized value for this stage.
   Q     - Quantization value. Represents desired quantization level size.
   state - State of quantizer.

   Return Value:
   -------------
   Inverse quantization value for this stage.

   Description:
   ------------
   initInvQuantSingleStage must be called prior to using this function the
   first time for a given value. It will compute the new, updated value
   based on the current index and state associated with the value.
*/

int invQuantSingleStage(int QIndex, int Q, quantState *state)
{
  int refLevs; /* how many refinement levels in new stage */
  int val;     /* new value to return for this stage */
  int sgn;

  /*--------------- INITIAL QUANTIZATION STAGE -------------------*/
  if (state->QR==0)
  {
    val = QIndex*Q + SGN(QIndex)*(QIndex ? QLEVMID(Q) : 0);

    /* update state */
    state->residualValue = QIndex*Q;
    state->QR            = Q;
    state->fromReduced   = 0;
    state->fromDeadZone  = (QIndex==0);

    return val;
  }


  /*--------------- NON-INITIAL QUANTIZATION STAGES -------------------*/

  /* get the number of refinement levels from lastQUsed state */
  refLevs = IROUND(state->QR,Q);

  /* Catch condition where there's no refinement being done.
     State information is not changed.
  */
  sgn = (state->residualValue < 0 || QIndex < 0) ? -1 : 1;      
  if (refLevs<=1)
    val = GETVAL(state->residualValue,sgn,state->QR);
  else
  {
    int inDeadZone;  /* are we still in the dead zone */
    int lastQUsed;   /* The "real" Q value last used */
    int lastLevSize; /* Size of quantization level used last */
    int newQUsed;
    int excess;
    int absQIndex;

    /* Initialize the last quant value used */
    lastQUsed = state->QR;
    
    /* update new Q value state */
    state->QR = newQUsed = ICEIL(lastQUsed,refLevs);
    
    /* Get last level size */
    lastLevSize = lastQUsed-(state->fromReduced);

    /* check if a reduced level can span the last level */
    if (refLevs*(newQUsed-1) >= lastLevSize)
    {
      --newQUsed;

      /* might overshoot (?) but can't reduce anymore */
      excess=0;
#if 1
      if (lastLevSize-refLevs*newQUsed)
	fprintf(stderr,"Excess in reduced partition\n");
#endif
    }
    else
      /* get excess (overshoot) */
      excess=lastLevSize-refLevs*newQUsed;

    /* Set dead zone indicator */
    inDeadZone = state->fromDeadZone;

    /*--- Calculate val. Update residualValue and fromReduced states ---*/
    absQIndex = abs(QIndex);
    if (excess==0)
    {
      state->fromReduced = (newQUsed < state->QR) ? 1 : 0;    
      state->residualValue += sgn*absQIndex*newQUsed;
    }
    else
    {
      int reducedIdx;
      int fullLevs;

      fullLevs = refLevs+excess;
      if (absQIndex >= fullLevs)
      {
       	state->fromReduced = 1;
	state->residualValue += sgn*fullLevs*newQUsed;

	--newQUsed;
	
	reducedIdx = absQIndex-fullLevs;
	state->residualValue += sgn*reducedIdx*newQUsed;
      }
      else
      {
	state->fromReduced = 0;
	state->residualValue += sgn*absQIndex*newQUsed;
      }
    }

    val = GETVAL(state->residualValue, sgn, newQUsed);
    
    if (inDeadZone && QIndex)
      state->fromDeadZone=0;
  }

  return val;
}


/* 
   Function:
   ---------
   invQuant - Complete inverse quantization of a value.

   Arguments:
   ----------
   int *vals - Pointer to array where de-quantized values are to be stored.
   int numQ - The number of quantization stages.
   int *QVals - Pointer to array where all the quantized values are stored.
   int *QList - The quantization values to be used for quantization. These
     will be modified if not multiples of one another.

   Return Value:
   -------------
   <none>

   Description:
   ------------
   invQuant will put the sequence of de-quantized values for the passed 
   sequence of quantized indices (values) value, QVals, in the array pointed
   to by vals. The inverse quantization is based on the numQ Q values in QList.
*/

void invQuant(int *vals, int numQ, int *QVals, int *QList)
{
  int curIdx;
  quantState state;

  /* initialize state */
  initInvQuantSingleStage(&state);
  for (curIdx = 0; curIdx < numQ; ++curIdx)
    vals[curIdx]=invQuantSingleStage(QVals[curIdx], QList[curIdx], &state);
}


/*--------- DERIVED QUANTIZATION VALUES AND REFINEMENT LEVELS ----------*/


/* 
   Function:
   ---------
   quantRefLev - Get the number of quantization levels for a given stage and
     the revised Q value.

   Arguments:
   ----------
   int curQ - Current input Q value.
   int *lastQUsed - last, revised, Q value. Will be updated.
   int whichQ - Quantization stage.

   Return Value:
   -------------
   The number of refinement quantization levels.

   Description:
   ------------
   quantRefLev will return the number of refinement quantization levels
   at stage whichQ based on the last Q value used and the current input Q
   value. It will also update the revised Q value by overwriting the argument
   lastQUsed.
*/

int quantRefLev(int curQ, int *lastQUsed, int whichQ)
{
  int refLevs;
  int newQUsed;

  /* get the number of refinement levels */
  refLevs = IROUND(*lastQUsed,curQ);
    
  if (whichQ==0 || refLevs>1)
  {
    /* get new level size */
    newQUsed = ICEIL(*lastQUsed,refLevs);

    /* update the last quant value used */
    *lastQUsed = newQUsed;
  }

  return refLevs;
}


/* 
   Function:
   ---------
   quantRevisedQList - Get the revised Q value list for all quantization
     stages.

   Arguments:
   ----------
   int *QList - Input list of Q values.
   int numQ - Number of Q values in QList (number of quantization stages).

   Return Value:
   -------------
   Pointer to allocated memory containing numQ revised Q values.

   Description:
   ------------
   quantRevisedQList takes in the input Q values list and returns the 
   revised list. Memory is allocated internally for the revised list.
*/

int *quantRevisedQList(int *QList, int numQ)
{
  int lastQUsed;
  int *revQList;
  int i;

  if ((revQList = (int *)malloc(numQ*sizeof(int))) != NULL)
  {
    lastQUsed=QList[0];
    for (i=0; i<numQ; ++i)
    {
      quantRefLev(QList[i], &lastQUsed, i);
      revQList[i]=lastQUsed;
    }
  }

  return revQList;
} 


/* 
   Function:
   ---------
   quantRefinementLevelList - Get the number of refinement levels for each
     of the quantization stages.

   Arguments:
   ----------
   int *QList - Input list of Q values.
   int numQ - Number of Q values in QList (number of quantization stages).

   Return Value:
   -------------
   Pointer to allocated memory containing numQ refinement level values.

   Description:
   ------------
   quantRefinementLevelList takes in the input Q values list and returns
   the number of refinement levels for each stage after revision. Memory is 
   allocated internally for the list.
*/

int *quantRefinementLevelList(int *QList, int numQ)
{
  int lastQUsed;
  int *refLevList;
  int i;

  if ((refLevList = (int *)malloc(numQ*sizeof(int))) != NULL)
  {
    lastQUsed=QList[0];
    for (i=0; i<numQ; ++i)
      refLevList[i]=quantRefLev(QList[i], &lastQUsed, i);
  }

  return refLevList;
} 


/*-------------- FOR TESTING AS STAND-ALONE PROGRAM ------------------*/

#if 0

/*--------------  Max/Min and Match Test for Range --------------*/

main()
{
  /* Example QLists:

    {70, 30, 1};    
    {16, 8, 4, 2, 1};
    {97, 43, 17, 1};
    {100, 20, 200, 100, 50, 25, 10, 5, 2, 1};
    {37, 25 , 18, 15, 12, 25, 16, 12, 11, 9, 9, 8, 7, 6, 5, 1};
    {140, 25, 5, 18, 9, 7, 6 , 5, 37, 25, 18, 15, 12,
      25, 16, 12, 11, 9, 9, 8, 7, 6, 5, 1};
    {81, 40, 20, 6, 1};
    {128, 64, 32, 16, 8, 4, 2, 1};
    {209, 105, 53, 27, 1};
  */
  int QList[]=    {100, 20, 200, 100, 50, 25, 10, 5, 2, 1};
  int *QVals;
  int val, *iVals;
  int numQ = sizeof(QList)/sizeof(int);
  int i;
  int *QValMax, *QValMin;
  int noMatchCnt;
  int *refLevList;
  int *revQList;
  int lastQUsed;

  iVals = (int *)malloc(sizeof(int)*numQ);
  QVals = (int *)malloc(sizeof(int)*numQ);
  QValMax = (int *)calloc(sizeof(int),numQ);
  QValMin = (int *)calloc(sizeof(int),numQ);
  refLevList = (int *)calloc(sizeof(int),numQ);
  revQList = (int *)calloc(sizeof(int),numQ);
  noMatchCnt=0;

  lastQUsed=QList[0];
  for (i=0; i<numQ; ++i)
  {
    refLevList[i] = quantRefLev(QList[i], &lastQUsed, i);
    revQList[i] = lastQUsed;
  }

  for (i=0; i<numQ; ++i)
    fprintf(stderr,"%d: QList=%d, revQList=%d, refLevList=%d\n",
	    i,QList[i],revQList[i],refLevList[i]);

  for (val= -4096; val <= 4096; ++val)
  {
    /* get val */
    quant(QVals, val, numQ, QList);
    
    for (i=0; i<numQ; ++i)
    {
      if (QValMax[i] < QVals[i])
	QValMax[i] = QVals[i];
      if (QValMin[i] > QVals[i])
	QValMin[i] = QVals[i];

      if (i && abs(QValMax[i]) >= refLevList[i])
	fprintf(stderr,"Error: val=%d, lev=%d, QValMax=%d >= refLevList=%d\n",
		val,i,QValMax[i],refLevList[i]);
      if (i && abs(QValMin[i]) >= refLevList[i])
	fprintf(stderr,"Error: val=%d, lev=%d, QValMin=%d >= refLevList=%d\n",
		val,i,QValMin[i],refLevList[i]);

      invQuant(iVals, numQ, QVals, QList);
    }
    if (iVals[numQ-1]!=val)
      ++noMatchCnt;
  }

  fprintf(stderr,"noMatchCnt=%d\n",noMatchCnt);
  for (i=0; i<numQ; ++i)
    fprintf(stderr,"%d: Max=%d, Min=%d\n",i, QValMax[i], QValMin[i]);

}

#endif




#if 0
/*-------------- User Input ------------------*/

main()
{
  /* Example QLists:

    {70, 30, 1};    
    {16, 8, 4, 2, 1};
    {97, 43, 17, 1};
    {100, 20, 200, 100, 50, 25, 10, 5, 2, 1};
    {97, 43, 17, 112, 40, 30, 69, 21, 8, 1};
  */

  int QList[]={140, 25, 5, 18, 9, 7, 6 , 5, 37, 25, 18, 15, 12,
      25, 16, 12, 11, 9, 9, 8, 7, 6, 5, 1};

  int *QVals;
  int val, *iVals;
  int numQ = sizeof(QList)/sizeof(int);
  int i;
  int *revQList;

  QVals = (int *)malloc(sizeof(int)*numQ);
  iVals = (int *)malloc(sizeof(int)*numQ);

  revQList=quantRevisedQList(QList, numQ);
  for (i=0; i<numQ; ++i)
    printf("%d: QList=%d, revQList=%d\n",i,QList[i],revQList[i]);
  

  for (;;)
  {
    /* get val */
    printf("Enter val => ");
    scanf("%d",&val);
    if (val == 0)
      exit(-1);
    printf("val is %d\n",val);

    /*------------  Quantization ----------------*/
    quant(QVals, val, numQ, QList);

    printf("Quantized values are:\n");
    for (i=0; i<numQ; ++i)
      printf("QVals[%d] (Q=%3d) = %d\n",i,QList[i],QVals[i]);

 
    /*----------- Inverse Quantization ------------------*/    
    invQuant(Vals, numQ, QVals, QList);

    printf("Inverse Quantized values are:\n");
    
    for (i=0; i<numQ; ++i)  
      printf("iVal[%d] (Q=%3d) = %d, percent diff = %f\n",
	     i,QList[i],iVals[i], 100.0*(val-iVals[i])/val);

    printf("\n");
  }
}

#endif
